%{
#define YYERROR_VERBOSE 1

#include "error.h"
#include "quad.h"

extern int lineno;
Quad_list ql;
%}

%token T_unit   "unit"
%token T_endu   "endu"
%token T_assign ":="
%token T_array  "array"
%token T_ne     "<>"
%token T_le     "<="
%token T_ge     ">="
%token T_ifb    "ifb"
%token T_jump   "jump"
%token T_label  "label"
%token T_jumpl  "jumpl"
%token T_call   "call"
%token T_par    "par"
%token T_ret    "ret"
%token T_dim    "dim"
%token T_eqeq   "=="
%token T_nxeq   "!="

%token T_RET    "RET"
%token T_result "$$"
%token T_obj    "obj"
%token T_fun    "fun"
%token T_temp   "temp"

%token T_true   "true"
%token T_false  "false"

%token T_bool   "bool"
%token T_char   "char"
%token T_float  "float"
%token T_int    "int"
%token T_of     "of"
%token T_ref    "ref"
%token T_arrow  "->"

%union {
  RepInt       n;
  RepFloat     f;
  RepChar      c;
  RepString    s;
  Identifier   id;
  Object       obj;
  Function     fun;
  Quad_list    quad_list;
  Quad         quad;
  Quad_opname  opname;
  Quad_operand operand;
  Quad_simple  simple;
  PassMode     passmode;
  Type         typ;
}

%token<id>  T_id
%token<n>   T_tempid
%token<n>   T_cint
%token<f>   T_cfloat
%token<c>   T_character
%token<s>   T_string

%token T_nl

%type<quad_list> quad_list;
%type<quad>      quad;
%type<n>         label;
%type<opname>    opname;
%type<operand>   operand;
%type<simple>    simple;
%type<simple>    constant;
%type<simple>    object;
%type<simple>    function;
%type<simple>    temporary;
%type<passmode>  pass_mode;
%type<typ>       typ;
%type<n>         offset;
%type<n>         stars;
%type<n>         more_stars;

%right      "->"
%nonassoc   "array" "of"
%left       "ref"

%%

program:
      quad_list               { ql = $1; }
   ;

quad_list:
      /* nothing */           { $$ = NULL; }
   |  T_nl quad_list          { $$ = $2; }
   |  quad T_nl quad_list     { $$ = quad_list($1, $3); }
   ;

quad:
      label ':' opname ',' operand ',' operand ',' operand
         { $$ = quad($1, $3, $5, $7, $9); }
   ;

label:
      T_cint           { $$ = $1; }
   ;

opname:
      "unit"           { $$ = quad_opcode_unit; }
   |  "endu"           { $$ = quad_opcode_endu; }
   |  '+'              { $$ = quad_opcode_plus; }
   |  '-'              { $$ = quad_opcode_minus; }
   |  '*'              { $$ = quad_opcode_times; }
   |  '/'              { $$ = quad_opcode_div; }
   |  '%'              { $$ = quad_opcode_mod; }
   |  ":="             { $$ = quad_opcode_assign; }
   |  "array"          { $$ = quad_opcode_array; }
   |  '='              { $$ = quad_opcode_eq; }
   |  "<>"             { $$ = quad_opcode_ne; }
   |  '<'              { $$ = quad_opcode_lt; }
   |  '>'              { $$ = quad_opcode_gt; }
   |  "<="             { $$ = quad_opcode_le; }
   |  ">="             { $$ = quad_opcode_ge; }
   |  "ifb"            { $$ = quad_opcode_ifb; }
   |  "jump"           { $$ = quad_opcode_jump; }
   |  "label"          { $$ = quad_opcode_label; }
   |  "jumpl"          { $$ = quad_opcode_jumpl; }
   |  "call"           { $$ = quad_opcode_call; }
   |  "par"            { $$ = quad_opcode_par; }
   |  "ret"            { $$ = quad_opcode_ret; }
   |  "dim"            { $$ = quad_opcode_dim; }
   |  "=="             { $$ = quad_opcode_eqeq; }
   |  "!="             { $$ = quad_opcode_nxeq; }
   ;

operand:
      simple           { $$ = quad_operand_simple($1); }
   |  '[' simple ']'   { $$ = quad_operand_deref($2); }
   |  '{' simple '}'   { $$ = quad_operand_addressof($2); }
   |  pass_mode        { $$ = quad_operand_passmode($1); }
   |  '-'              { $$ = quad_operand_empty(); }
   ;

simple:
      constant         { $$ = $1; }
   |  object           { $$ = $1; }
   |  function         { $$ = $1; }
   |  temporary        { $$ = $1; }
   |  "$$"             { $$ = quad_result(); }

constant:
      T_cint           { $$ = quad_iconst($1); }
   |  T_cfloat         { $$ = quad_fconst($1); }
   |  T_character      { $$ = quad_cconst($1); }
   |  T_string         { $$ = quad_strlit($1); }
   |  "true"           { $$ = quad_true(); }
   |  "false"          { $$ = quad_false(); }
   ;

offset:
      T_cint           { $$ = $1; }
   |  '+' T_cint       { $$ = $2; }
   |  '-' T_cint       { $$ = - $2; }
   ;

object:
      "obj" '[' T_id ';' typ ';' offset ';' T_cint ';' T_cint ']'
        { Object obj;
	  obj.id = $3;
	  obj.typ = $5;
	  obj.offset = $7;
	  obj.nesting = $9;
	  obj.byref = $11 != 0;
	  $$ = quad_object(obj); }
   ;

function:
      "fun" '[' T_id ';' T_cint ';' T_cint ';' T_cint ';' T_cint ';' T_cint ']'
        { Function fun;
	  fun.id = $3;
	  fun.unique = $5;
	  fun.paramBytes = $7;
	  fun.localBytes = $9;
	  fun.nesting = $11;
	  fun.hasResult = $13 != 0;
	  $$ = quad_function(fun); }
   ;

typ:
      "unit"                     { $$ = type_unit(); }
   |  "int"                      { $$ = type_int(); }
   |  "char"                     { $$ = type_char(); }
   |  "bool"                     { $$ = type_bool(); }
   |  "float"                    { $$ = type_float(); }
   |  '(' typ ')'                { $$ = $2; }
   |  typ "->" typ               { $$ = type_func($1, $3); }
   |  typ "ref"                  { $$ = type_ref($1); }
   |  "array" stars "of" typ     { $$ = type_array($2, $4); }
   |  T_id                       { $$ = type_id($1); }
   ;

stars:
      /*nothing*/                { $$ = 1; }
   |  '[' '*' more_stars ']'     { $$ = 1 + $3; }
   ;

more_stars:
      /*nothing*/                { $$ = 0; }
   |  ',' '*' more_stars         { $$ = 1 + $3; }
   ;

temporary:
      "temp" '[' T_tempid ';' typ ';' offset ']'
        { Temporary temp;
	  temp.num = $3;
	  temp.typ = $5;
	  temp.offset = $7;
	  $$ = quad_temporary(temp); }
   ;

pass_mode:
      'V'              { $$ = pass_byval; }
   |  'R'              { $$ = pass_byref; }
   |  "RET"            { $$ = pass_ret; }
   ;

%%

int yyerror (const char *msg)
{
   error("syntax error: %s" , msg);
}

int main ()
{
   int result;

   filename="stdin";
   linecount=1;
   result=yyparse();
   if (result==0) {
      printSimple = true;
      quad_list_print(stdout, ql);
   }
   exit(0);
}
